/*
Turing Machine Simulator
Copyright (C) 2005-2011 Samuel BWPOW Kupka, http://bwpow.eu

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
#include <cstdio>
#include <cstring>
#include <string>
#include <vector>
#include <map>
#include <algorithm>
#include "ts.h"

using namespace std;

extern map<INSTR,INSTR> instr;
extern vector<string> state;
extern vector<char> pr,pl;
extern INFO info;
extern char chars[256];

int run_get_p(int pos)
{
  if(pos>=0){
    if(info.cr&&pos>=info.maxmemory){
      if(info.nt) return '<';
      return -1;
    }
    if(pos>=(int)pr.size()) return info.zero;
    return pr[pos];
  }
  else{
    pos=abs(pos)-1;
    if(info.cl){
      if(info.nt) return '>';
      return -1;
    }
    if(pos>=(int)pl.size()) return info.zero;
    return pl[pos];
  }
  return info.zero;
}

void run_write_p(int pos,int c)
{
  if(pos>=0){
    if(info.cr){
      if(pos>=info.maxmemory) return;
    }
    if(pos>=(int)pr.size()+1) return;
    if(pos==(int)pr.size()){pr.push_back(c); return; }
    pr[pos]=c;
  }
  else{
    pos=abs(pos)-1;
    if(info.cl) return;
    if(pos>=(int)pl.size()+1) return;
    if(pos==(int)pl.size()){pl.push_back(c); return; }
    pl[pos]=c;
  }
}

int run(void)
{
  int pos=0;
  INSTR a,b;
  a.m=0;
  info.state=0;
  info.steps=0;

  if(info.spos==1){
    for(pos=0;pos<(int)pr.size();pos++){
        if(pr[pos]==info.one) break;
    }
    if(pos==(int)pr.size()){
        pos=0;
    }
  }

  while(1){
    if(info.steps>info.maxsteps) return -1;

    a.s=info.state;
    a.p=run_get_p(pos);
    info.pos=pos;

    if(instr.count(a)<=0) break;
    b=instr[a];

    run_write_p(pos,b.p);
    info.state=b.s;

    if(b.m==MOVE_LEFT) pos--;
    if(b.m==MOVE_RIGHT) pos++;

    if((int)(pr.size()+pl.size())>info.maxmemory) return -2;
    info.steps++;
  }
  return 0;
}
